@@ -40,7 +40,9 @@
 #include "qquickmousearea_p.h"
 #include "qquickmousearea_p_p.h"
 #include "qquickwindow.h"
+#if QT_CONFIG(quick_draganddrop)
 #include "qquickdrag_p.h"
+#endif
 
 #include <private/qqmldata_p.h>
 #include <private/qsgadaptationlayer_p.h>
@@ -55,25 +57,25 @@ QT_BEGIN_NAMESPACE
 
 DEFINE_BOOL_CONFIG_OPTION(qmlVisualTouchDebugging, QML_VISUAL_TOUCH_DEBUGGING)
 
-Q_DECLARE_LOGGING_CATEGORY(DBG_HOVER_TRACE)
+Q_DECLARE_LOGGING_CATEGORY(lcHoverTrace)
 
 QQuickMouseAreaPrivate::QQuickMouseAreaPrivate()
 : enabled(true), scrollGestureEnabled(true), hovered(false), longPress(false),
   moved(false), stealMouse(false), doubleClick(false), preventStealing(false),
-  propagateComposedEvents(false), overThreshold(false), pressed(0),
+  propagateComposedEvents(false), overThreshold(false),
   pressAndHoldInterval(-1)
-#if QT_CONFIG(draganddrop)
-  , drag(0)
+#if QT_CONFIG(quick_draganddrop)
+  , drag(nullptr)
 #endif
 #if QT_CONFIG(cursor)
-  , cursor(0)
+  , cursor(nullptr)
 #endif
 {
 }
 
 QQuickMouseAreaPrivate::~QQuickMouseAreaPrivate()
 {
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
     delete drag;
 #endif
 #if QT_CONFIG(cursor)
@@ -94,11 +96,12 @@ void QQuickMouseAreaPrivate::init()
 
 void QQuickMouseAreaPrivate::saveEvent(QMouseEvent *event)
 {
-    lastPos = event->localPos();
-    lastScenePos = event->windowPos();
+    lastPos = event->position();
+    lastScenePos = event->scenePosition();
     lastButton = event->button();
     lastButtons = event->buttons();
     lastModifiers = event->modifiers();
+    lastFlags = event->flags();
 }
 
 bool QQuickMouseAreaPrivate::isPressAndHoldConnected()
@@ -195,7 +198,7 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     \instantiates QQuickMouseArea
     \inqmlmodule QtQuick
     \ingroup qtquick-input
-    \brief Enables simple mouse handling
+    \brief Enables simple mouse handling.
     \inherits Item
 
     A MouseArea is an invisible item that is typically used in conjunction with
@@ -272,8 +275,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     even when no mouse button is pressed.
 
     \sa hoverEnabled
-
-    The corresponding handler is \c onEntered.
 */
 
 /*!
@@ -312,8 +313,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     both be considered to be simultaneously hovered.
 
     \sa hoverEnabled
-
-    The corresponding handler is \c onExited.
 */
 
 /*!
@@ -330,8 +329,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
 
     When handling this signal, changing the \l {MouseEvent::}{accepted} property of the \a mouse
     parameter has no effect.
-
-    The corresponding handler is \c onPositionChanged.
 */
 
 /*!
@@ -346,8 +343,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
 
     When handling this signal, changing the \l {MouseEvent::}{accepted} property of the \a mouse
     parameter has no effect, unless the \l propagateComposedEvents property is \c true.
-
-    The corresponding handler is \c onClicked.
 */
 
 /*!
@@ -362,8 +357,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     release. The default is to accept the event and not allow other MouseAreas beneath this one to
     handle the event.  If \e accepted is set to false, no further events will be sent to this MouseArea
     until the button is next pressed.
-
-    The corresponding handler is \c onPressed.
 */
 
 /*!
@@ -376,8 +369,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     When handling this signal, changing the \l {MouseEvent::}{accepted} property of the \a mouse
     parameter has no effect.
 
-    The corresponding handler is \c onReleased.
-
     \sa canceled
 */
 
@@ -390,8 +381,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
 
     When handling this signal, changing the \l {MouseEvent::}{accepted} property of the \a mouse
     parameter has no effect, unless the \l propagateComposedEvents property is \c true.
-
-    The corresponding handler is \c onPressAndHold.
 */
 
 /*!
@@ -404,8 +393,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     When handling this signal, if the \l {MouseEvent::}{accepted} property of the \a mouse
     parameter is set to false, the pressed/released/clicked signals will be emitted for the second
     click; otherwise they are suppressed.  The \c accepted property defaults to true.
-
-    The corresponding handler is \c onDoubleClicked.
 */
 
 /*!
@@ -419,8 +406,6 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
     \l Flickable will steal the mouse handling from the MouseArea. In these cases, to reset
     the logic when the MouseArea has lost the mouse handling to the \l Flickable,
     \c canceled should be handled in addition to \l released.
-
-    The corresponding handler is \c onCanceled.
 */
 
 /*!
@@ -428,11 +413,9 @@ bool QQuickMouseAreaPrivate::propagateHelper(QQuickMouseEvent *ev, QQuickItem *i
 
     This signal is emitted in response to both mouse wheel and trackpad scroll gestures.
 
-    The \l {WheelEvent}{wheel} parameter provides information about the event, including the x and y
+    The \a wheel parameter provides information about the event, including the x and y
     position, any buttons currently pressed, and information about the wheel movement, including
     angleDelta and pixelDelta.
-
-    The corresponding handler is \c onWheel.
 */
 
 QQuickMouseArea::QQuickMouseArea(QQuickItem *parent)
@@ -614,7 +597,7 @@ void QQuickMouseArea::setPreventStealing(bool prevent)
             MouseArea {
                 anchors.fill: parent
                 propagateComposedEvents: true
-                onClicked: {
+                onClicked: (mouse)=> {
                     console.log("clicked blue")
                     mouse.accepted = false
                 }
@@ -684,12 +667,12 @@ void QQuickMouseArea::mousePressEvent(QMouseEvent *event)
     } else {
         d->longPress = false;
         d->saveEvent(event);
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
         if (d->drag)
             d->drag->setActive(false);
 #endif
         setHovered(true);
-        d->startScene = event->windowPos();
+        d->startScene = event->scenePosition();
         setKeepMouseGrab(d->stealMouse);
         event->setAccepted(setPressed(event->button(), true, event->source()));
         if (event->isAccepted())
@@ -705,13 +688,19 @@ void QQuickMouseArea::mouseMoveEvent(QMouseEvent *event)
         return;
     }
 
-    d->saveEvent(event);
-
     // ### we should skip this if these signals aren't used
     // ### can GV handle this for us?
-    setHovered(contains(d->lastPos));
+    setHovered(contains(event->position()));
 
-#if QT_CONFIG(draganddrop)
+    if ((event->buttons() & acceptedMouseButtons()) == 0) {
+        QQuickItem::mouseMoveEvent(event);
+        return;
+    }
+
+    d->saveEvent(event);
+
+
+#if QT_CONFIG(quick_draganddrop)
     if (d->drag && d->drag->target()) {
         if (!d->moved) {
             d->targetStartPos = d->drag->target()->parentItem()
@@ -723,10 +712,10 @@ void QQuickMouseArea::mouseMoveEvent(QMouseEvent *event)
         QPointF curLocalPos;
         if (drag()->target()->parentItem()) {
             startLocalPos = drag()->target()->parentItem()->mapFromScene(d->startScene);
-            curLocalPos = drag()->target()->parentItem()->mapFromScene(event->windowPos());
+            curLocalPos = drag()->target()->parentItem()->mapFromScene(event->scenePosition());
         } else {
             startLocalPos = d->startScene;
-            curLocalPos = event->windowPos();
+            curLocalPos = event->scenePosition();
         }
 
         if (keepMouseGrab() && d->stealMouse && d->overThreshold && !d->drag->active())
@@ -758,20 +747,22 @@ void QQuickMouseArea::mouseMoveEvent(QMouseEvent *event)
 
         QPointF targetPos = d->drag->target()->position();
 
-        if (d->drag->active())
+        if (d->drag->active()) {
             d->drag->target()->setPosition(boundedDragPos);
+            d->lastPos = mapFromScene(d->lastScenePos);
+        }
 
-        bool dragOverThresholdX = QQuickWindowPrivate::dragOverThreshold(dragPos.x() - startPos.x(),
-                                                                         Qt::XAxis, event, d->drag->threshold());
-        bool dragOverThresholdY = QQuickWindowPrivate::dragOverThreshold(dragPos.y() - startPos.y(),
-                                                                         Qt::YAxis, event, d->drag->threshold());
+        bool dragOverThresholdX = QQuickDeliveryAgentPrivate::dragOverThreshold(dragPos.x() - startPos.x(),
+                                                                                Qt::XAxis, event, d->drag->threshold());
+        bool dragOverThresholdY = QQuickDeliveryAgentPrivate::dragOverThreshold(dragPos.y() - startPos.y(),
+                                                                                Qt::YAxis, event, d->drag->threshold());
 
         if (!d->overThreshold && (((targetPos.x() != boundedDragPos.x()) && dragOverThresholdX) ||
                                   ((targetPos.y() != boundedDragPos.y()) && dragOverThresholdY)))
         {
             d->overThreshold = true;
             if (d->drag->smoothed())
-                d->startScene = event->windowPos();
+                d->startScene = event->scenePosition();
         }
 
         if (!keepMouseGrab() && d->overThreshold) {
@@ -784,7 +775,7 @@ void QQuickMouseArea::mouseMoveEvent(QMouseEvent *event)
 #endif
 
     QQuickMouseEvent &me = d->quickMouseEvent;
-    me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, false, d->longPress);
+    me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, false, d->longPress, event->flags());
     me.setSource(event->source());
     emit mouseXChanged(&me);
     me.setPosition(d->lastPos);
@@ -805,7 +796,7 @@ void QQuickMouseArea::mouseReleaseEvent(QMouseEvent *event)
         setPressed(event->button(), false, event->source());
         if (!d->pressed) {
             // no other buttons are pressed
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
             if (d->drag)
                 d->drag->setActive(false);
 #endif
@@ -827,7 +818,8 @@ void QQuickMouseArea::mouseDoubleClickEvent(QMouseEvent *event)
     if (d->enabled) {
         d->saveEvent(event);
         QQuickMouseEvent &me = d->quickMouseEvent;
-        me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, true, false);
+        me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, true,
+                 false, event->flags());
         me.setSource(event->source());
         me.setAccepted(d->isDoubleClickConnected());
         emit this->doubleClicked(&me);
@@ -844,7 +836,7 @@ void QQuickMouseArea::hoverEnterEvent(QHoverEvent *event)
     if (!d->enabled && !d->pressed) {
         QQuickItem::hoverEnterEvent(event);
     } else {
-        d->lastPos = event->posF();
+        d->lastPos = event->position();
         d->lastModifiers = event->modifiers();
         setHovered(true);
         QQuickMouseEvent &me = d->quickMouseEvent;
@@ -861,8 +853,8 @@ void QQuickMouseArea::hoverMoveEvent(QHoverEvent *event)
     Q_D(QQuickMouseArea);
     if (!d->enabled && !d->pressed) {
         QQuickItem::hoverMoveEvent(event);
-    } else if (d->lastPos != event->posF()) {
-        d->lastPos = event->posF();
+    } else if (d->lastPos != event->position()) {
+        d->lastPos = event->position();
         d->lastModifiers = event->modifiers();
         QQuickMouseEvent &me = d->quickMouseEvent;
         me.reset(d->lastPos.x(), d->lastPos.y(), Qt::NoButton, Qt::NoButton, d->lastModifiers, false, false);
@@ -893,8 +885,7 @@ void QQuickMouseArea::wheelEvent(QWheelEvent *event)
     }
 
     QQuickWheelEvent &we = d->quickWheelEvent;
-    we.reset(event->posF().x(), event->posF().y(), event->angleDelta(), event->pixelDelta(),
-             event->buttons(), event->modifiers(), event->inverted());
+    we.reset(event);
     we.setAccepted(d->isWheelConnected());
     emit wheel(&we);
     if (!we.isAccepted())
@@ -908,13 +899,13 @@ void QQuickMouseArea::ungrabMouse()
     if (d->pressed) {
         // if our mouse grab has been removed (probably by Flickable), fix our
         // state
-        d->pressed = 0;
+        d->pressed = Qt::NoButton;
         d->stealMouse = false;
         d->doubleClick = false;
         d->overThreshold = false;
         setKeepMouseGrab(false);
 
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
         if (d->drag)
             d->drag->setActive(false);
 #endif
@@ -936,16 +927,22 @@ void QQuickMouseArea::mouseUngrabEvent()
     ungrabMouse();
 }
 
+void QQuickMouseArea::touchUngrabEvent()
+{
+    // allow a Pointer Handler to steal the grab from MouseArea
+    ungrabMouse();
+}
+
 bool QQuickMouseArea::sendMouseEvent(QMouseEvent *event)
 {
     Q_D(QQuickMouseArea);
-    QPointF localPos = mapFromScene(event->windowPos());
+    QPointF localPos = mapFromScene(event->scenePosition());
 
     QQuickWindow *c = window();
-    QQuickItem *grabber = c ? c->mouseGrabberItem() : 0;
+    QQuickItem *grabber = c ? c->mouseGrabberItem() : nullptr;
     bool stealThisEvent = d->stealMouse;
     if ((stealThisEvent || contains(localPos)) && (!grabber || !grabber->keepMouseGrab())) {
-        QMouseEvent mouseEvent(event->type(), localPos, event->windowPos(), event->screenPos(),
+        QMouseEvent mouseEvent(event->type(), localPos, event->scenePosition(), event->globalPosition(),
                                event->button(), event->buttons(), event->modifiers());
         mouseEvent.setAccepted(false);
 
@@ -963,7 +960,7 @@ bool QQuickMouseArea::sendMouseEvent(QMouseEvent *event)
         default:
             break;
         }
-        grabber = c ? c->mouseGrabberItem() : 0;
+        grabber = c ? c->mouseGrabberItem() : nullptr;
         if (grabber && stealThisEvent && !grabber->keepMouseGrab() && grabber != this)
             grabMouse();
 
@@ -997,7 +994,7 @@ bool QQuickMouseArea::childMouseEventFilter(QQuickItem *i, QEvent *e)
     Q_D(QQuickMouseArea);
     if (!d->pressed &&
             (!d->enabled || !isVisible()
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
              || !d->drag || !d->drag->filterChildren()
 #endif
             )
@@ -1020,7 +1017,7 @@ void QQuickMouseArea::timerEvent(QTimerEvent *event)
     Q_D(QQuickMouseArea);
     if (event->timerId() == d->pressAndHoldTimer.timerId()) {
         d->pressAndHoldTimer.stop();
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
         bool dragged = d->drag && d->drag->active();
 #else
         bool dragged = false;
@@ -1028,7 +1025,7 @@ void QQuickMouseArea::timerEvent(QTimerEvent *event)
         if (d->pressed && dragged == false && d->hovered == true) {
             d->longPress = true;
             QQuickMouseEvent &me = d->quickMouseEvent;
-            me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, false, d->longPress);
+            me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, false, d->longPress, d->lastFlags);
             me.setSource(Qt::MouseEventSynthesizedByQt);
             me.setAccepted(d->isPressAndHoldConnected());
             emit pressAndHold(&me);
@@ -1040,17 +1037,10 @@ void QQuickMouseArea::timerEvent(QTimerEvent *event)
     }
 }
 
-void QQuickMouseArea::windowDeactivateEvent()
-{
-    ungrabMouse();
-    QQuickItem::windowDeactivateEvent();
-}
-
-void QQuickMouseArea::geometryChanged(const QRectF &newGeometry,
-                                            const QRectF &oldGeometry)
+void QQuickMouseArea::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
     Q_D(QQuickMouseArea);
-    QQuickItem::geometryChanged(newGeometry, oldGeometry);
+    QQuickItem::geometryChange(newGeometry, oldGeometry);
 
     if (d->lastScenePos.isNull)
         d->lastScenePos = mapToScene(d->lastPos);
@@ -1063,7 +1053,7 @@ void QQuickMouseArea::itemChange(ItemChange change, const ItemChangeData &value)
     Q_D(QQuickMouseArea);
     switch (change) {
     case ItemVisibleHasChanged:
-        if (acceptHoverEvents() && d->hovered != (isVisible() && isUnderMouse())) {
+        if (d->effectiveEnable && d->enabled && acceptHoverEvents() && d->hovered != (isVisible() && isUnderMouse())) {
             if (!d->hovered) {
                 QPointF cursorPos = QGuiApplicationPrivate::lastCursorPosition;
                 d->lastScenePos = d->window->mapFromGlobal(cursorPos.toPoint());
@@ -1071,6 +1061,12 @@ void QQuickMouseArea::itemChange(ItemChange change, const ItemChangeData &value)
             }
             setHovered(!d->hovered);
         }
+        if (d->pressed && (!isVisible())) {
+            // This happens when the mouse area sets itself disabled or hidden
+            // inside the press handler. In that case we should not keep the internal
+            // state as pressed, since we never became the mouse grabber.
+            ungrabMouse();
+        }
         break;
     default:
         break;
@@ -1150,7 +1146,7 @@ void QQuickMouseArea::setHovered(bool h)
 {
     Q_D(QQuickMouseArea);
     if (d->hovered != h) {
-        qCDebug(DBG_HOVER_TRACE) << this << d->hovered <<  "->" << h;
+        qCDebug(lcHoverTrace) << this << d->hovered <<  "->" << h;
         d->hovered = h;
         emit hoveredChanged();
         d->hovered ? emit entered() : emit exited();
@@ -1196,7 +1192,7 @@ bool QQuickMouseArea::setPressed(Qt::MouseButton button, bool p, Qt::MouseEventS
 {
     Q_D(QQuickMouseArea);
 
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
     bool dragged = d->drag && d->drag->active();
 #else
     bool dragged = false;
@@ -1207,7 +1203,7 @@ bool QQuickMouseArea::setPressed(Qt::MouseButton button, bool p, Qt::MouseEventS
 
     if (wasPressed != p) {
         QQuickMouseEvent &me = d->quickMouseEvent;
-        me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, isclick, d->longPress);
+        me.reset(d->lastPos.x(), d->lastPos.y(), d->lastButton, d->lastButtons, d->lastModifiers, isclick, d->longPress, d->lastFlags);
         me.setSource(source);
         if (p) {
             d->pressed |= button;
@@ -1364,6 +1360,7 @@ void QQuickMouseArea::resetPressAndHoldInterval()
     \qmlproperty real QtQuick::MouseArea::drag.maximumY
     \qmlproperty bool QtQuick::MouseArea::drag.filterChildren
     \qmlproperty real QtQuick::MouseArea::drag.threshold
+    \qmlproperty bool QtQuick::MouseArea::drag.smoothed
 
     \c drag provides a convenient way to make an item draggable.
 
@@ -1402,7 +1399,7 @@ void QQuickMouseArea::resetPressAndHoldInterval()
 
 */
 
-#if QT_CONFIG(draganddrop)
+#if QT_CONFIG(quick_draganddrop)
 QQuickDrag *QQuickMouseArea::drag()
 {
     Q_D(QQuickMouseArea);
@@ -1418,7 +1415,7 @@ QSGNode *QQuickMouseArea::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeData
     Q_D(QQuickMouseArea);
 
     if (!qmlVisualTouchDebugging())
-        return 0;
+        return nullptr;
 
     QSGInternalRectangleNode *rectangle = static_cast<QSGInternalRectangleNode *>(oldNode);
     if (!rectangle) rectangle = d->sceneGraphContext()->createInternalRectangleNode();
